############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import time, types, warnings

import ZSI
from ZSI import EvaluateException, ParseException
from ZSI.parse import ParsedSoap
from ZSI.fault import FaultFromFaultMessage, FaultType
from ZSI.address import Address, WSActionException
from ZSI.writer import SoapWriter
from ZSI.wstools.Namespaces import OASIS

from twisted.web import client
from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log, failure
from twisted.python.failure import Failure
from twisted.internet.tcp import *

from pyGridWare.faults import WrapFaultInException
try:
    from pyGridWare.utility.dom import DomletteReader, DomletteElementProxy
except ImportError, ex:
    DomletteReader = DomletteElementProxy = None


class HTTPPageGetter(client.HTTPPageGetter):
    def handleStatus_500(self):
        '''potentially a SOAP:Fault.
        '''
        pass

try:
    from OpenSSL import SSL
    SSL.Session
except:
    client.HTTPClientFactory.protocol = HTTPPageGetter
else:
    class SMHTTPPageGetter(HTTPPageGetter):
        """Retains the SSL session
        """
        session = None

        def makeConnection(self, transport):
            """Do handshake and set session id.
            """
            socket = transport.getHandle()
            use_session = hasattr(socket, 'do_handshake')
            session = SMHTTPPageGetter.session
            if use_session is True:
                if session is None:
                    # JRB Must call do_handshake, but sometimes 
                    # getting WantReadError??
                    #try:
                    #    socket.do_handshake()
                    #except SSL.WantReadError, ex: 
                    #    warnings.warn('Got WantReadError', stacklevel=2)
                    SMHTTPPageGetter.session = SSL.Session(socket)
                else:
                    socket.set_session(session)

            client.HTTPPageGetter.makeConnection(self, transport)

    client.HTTPClientFactory.protocol = SMHTTPPageGetter
    


def getPage(url, contextFactory=None, *args, **kwargs):
    """Download a web page as a string.

    Download a page. Return a deferred, which will callback with a
    page (as a string) or errback with a description of the error.

    See HTTPClientFactory to see what extra args can be passed.
    """
    scheme, host, port, path = client._parse(url)
    factory = client.HTTPClientFactory(url, *args, **kwargs)
    if scheme == 'https':
        if contextFactory is None:
            raise RuntimeError, 'must provide a contextFactory'
        conn = reactor.connectSSL(host, port, factory, contextFactory)
    else:
        conn = reactor.connectTCP(host, port, factory)

    return factory


class Binding:
    def __init__(self, url=None, nsdict=None, readerclass=DomletteReader, 
                 writerclass=DomletteElementProxy, wsAddressURI=None, 
                 sig_handler=None, contextFactory=None, **kw):

        self.url = url
        self.nsdict = nsdict or {}
        self.readerclass = readerclass
        self.writerclass = writerclass
        self.wsAddressURI = wsAddressURI
        self.sig_handler = sig_handler
        self.endPointReference = kw.get('endPointReference')
        self.contextFactory = contextFactory

        self._factory = None
        self._agent = 'pyGridWare Binding'
        self.http_headers  = {'content-type': 'text/xml',}
        self._fault = False

    def addHTTPHeader(self, key, value):
        self.http_headers[key] = value
   
    def _getHTTPHeaders(self):
        return self.http_headers
        
    def Send(self, url, opname, obj, nsdict={}, soapaction=None, **kw):
        '''
        '''
        url = url or self.url
        cookies = None
        if self._factory is not None:
            cookies = self._factory.cookies
         
        self.addHTTPHeader('SOAPAction', soapaction)
        soapdata = self._getRequestSOAPMessage(url, obj, nsdict, soapaction, **kw)
        self._factory = getPage(str(url), contextFactory=self.contextFactory, 
                          postdata=soapdata, agent=self._agent, 
                          method='POST', headers=self._getHTTPHeaders(), cookies=cookies)
        self._defer = d = self._factory.deferred
        return d

    def Receive(self, replytype, **kw):
        '''
        '''
        assert getattr(self, '_defer') is not None, 'Must call Send before Receive.'
        assert isinstance(replytype, ZSI.TC.TypeCode), 'replytype must be a TypeCode.'

        self.pyobj = None
        self._defer.addCallback(self._getResponseSOAPMessage, replytype=replytype, **kw)

        # 
        # Go from asynch --> synchronous
        # Cannot prevent other calls from being queued, 
        # only waiting for deferred to fire.
        # 
        while not self._defer.called:
            reactor.runUntilCurrent()
            t2 = reactor.timeout()
            t = reactor.running and t2
            reactor.doIteration(t)

        pyobj = self._defer.result
        if isinstance(pyobj, Exception) is True:
            raise pyobj

        return pyobj

    def _getResponseSOAPMessage(self, soapdata, replytype, **kw):
        '''
        '''
        if len(soapdata) == 0:
            raise TypeError('Received empty response')

        log.debug("_" * 33, time.ctime(time.time()), "RESPONSE: \n%s" %soapdata)

        ps = ParsedSoap(soapdata, readerclass=self.readerclass)
        if ps.IsAFault() is True:
            return FaultFromFaultMessage(ps) 

        self.pyobj = pyobj = ps.Parse(replytype)
        address = self.address 
        if address is not None:
            wsaction = kw.get('wsaction', None)
            if wsaction is not None:
                address.checkResponse(ps, wsaction)

        if self.sig_handler is not None:
            self.sig_handler.verify(ps)

        return pyobj

    def _getRequestSOAPMessage(self, url, obj, nsdict={}, soapaction=None, **kw):
        '''returns string representing request
        '''
        url = url or self.url
        if kw.has_key('requesttypecode'):
            tc = kw['requesttypecode']
        elif kw.has_key('requestclass'):
            tc = kw['requestclass'].typecode
        elif type(obj) == types.InstanceType:
            tc = getattr(obj.__class__, 'typecode')
        elif hasattr(obj, 'typecode'):
            tc = getattr(obj.__class__, 'typecode')
        else:
            raise RuntimeError, 'Must provide a typecode to serialize obj'

        # Serialize the object.
        d = {}
        d.update(self.nsdict)
        d.update(nsdict)

        use_header = self.sig_handler is not None or self.wsAddressURI is not None
        sw = SoapWriter(nsdict=d, header=use_header, outputclass=self.writerclass)
        sw.serialize(obj, tc)

        # Serialize WS-Address
        address = None
        if self.wsAddressURI:
            epr = kw.get('endPointReference', self.endPointReference)
            wsaction = kw.get('wsaction')
            if soapaction and wsaction.strip('\'"') != soapaction:
                raise WSActionException, 'soapAction(%s) and WS-Action(%s) must match'\
                    %(soapaction,wsaction)

            address = Address(url, wsAddressURI=self.wsAddressURI)
            address.setRequest(epr, wsaction)
            address.serialize(sw)
        self.address = address

        if self.sig_handler is not None:
            self.sig_handler.sign(sw)
        soapdata = str(sw)

        log.debug("_" * 33, time.ctime(time.time()), "REQUEST: \n%s" %soapdata)

        return soapdata


