##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, warnings
from os.path import isfile

from ZSI.wstools import WSDLTools
from ZSI.wstools.Namespaces import WSRF_V1_2
from ZSI.generate.utility import TextProtect, NCName_to_ClassName as NC_to_CN

"""
Functions:
    GetResourcePropertiesFromPort(port):
    GetContextClassBasesFromPort(port):
    GetPropertiesBasesFromRP(ged):

Classes:
    PortTypeMap:
    NameResolver:
"""

HEADER = """############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
# Generated by setup.py
###########################################################################
"""

CLIENT = 'client'
SERVICE = 'service'
SERVICE_STUBS = 'service_stubs'
TYPES = 'types'
PROPERTIES = 'properties'
RESOURCE = 'resource'
GEN_PACKAGES = None

def GetResourcePropertiesFromPort(port):
    '''Resolves ResourceProperty declaration's element references
    and returns all element declarations.
    '''
    els = []
    wsdl = port.getWSDL()
    pt = port.getPortType()
    rp = pt.getResourceProperties()
    if rp is None:
        warnings.warn(\
            'portType(%s) does not have a ResourceProperties declaration' \
             %(pt.name),
        )
        return None

    nsuri,name = rp
    schema = wsdl.types.get(nsuri)
    assert schema is not None and schema.elements.has_key(name) is True, \
        'portType(%s) ResourcePropties declares undefined element(%s,%s)'\
        %(pt.name, nsuri, name)

    ged = schema.elements[name]
    typeDef = ged.getTypeDefinition()
    assert typeDef.isComplex(), \
        'resource properties type definition must define a sequence or all model group: %s'\
        %(typeDef.getItemTrace())

    assert typeDef.content is not None \
        or not typeDef.content.isSequence() or not typeDef.content.isAll(),\
        'resource properties type definition must define a sequence or all model group: %s'\
        %(typeDef.getItemTrace())

    for er in typeDef.content.content:
        assert er.isReference() and er.isElement(), \
            'ReferenceProperties(%s) must specify element references..'\
            %ged.getItemTrace()

        e = er.getElementDeclaration()
        assert e.isDeclaration() and e.isElement(), \
            'WSDL Definition(%s) no element(%s,%s) for element reference: %s'\
            %(wsdl.name, er.getTargetNamespace(), er.getAttribute('ref'), ged.getItemTrace())

        els.append(e)

    assert ged is not None, \
        'Must declare ResourceProperties in PortType, from WSDL Definition(%s)'\
        %(wsdl.name)

    return els 


def GetContextClassBasesFromPort(port):
    '''Examines port's portType operation's input messages 
    for WSRF specification messages.  Assume output messages
    are corrrect..
    '''
    keys = []
    bases = []
    properties = lifetime = base_notification = None

    keys = map(lambda e: (e.getTargetNamespace(),e.getAttribute('name')),
               GetResourcePropertiesFromPort(port))
    for prop in keys:
        nsuri,name = prop
        if nsuri in WSRF_V1_2.LIFETIME.XSD_LIST:
            lifetime = nsuri
        elif nsuri in WSRF_V1_2.PROPERTIES.XSD_LIST:
            properties = nsuri
        elif nsuri in WSRF_V1_2.BASENOTIFICATION.XSD_LIST:
            base_notification = nsuri
   
    if lifetime is not None and base_notification is not None: 
        p = [\
            (lifetime,'CurrentTime'),
            (lifetime,'TerminationTime'),
            (base_notification,'ConsumerReference'),
            (base_notification,'TopicExpression'),
            (base_notification,'UseNotify'),
            (base_notification,'Precondition'),
            (base_notification,'Selector'),
            (base_notification,'SubscriptionPolicy'),
            (base_notification,'CreationTime'),
        ]
        for key in p:
            if key not in keys: break
        else: bases.append(PortTypeMap.context(PortTypeMap.SubscriptionManager))

    if lifetime is not None and PortTypeMap.context(PortTypeMap.SubscriptionManager) not in bases:
        p = [\
            (lifetime,'CurrentTime'),
            (lifetime,'TerminationTime'),
        ]
        for key in p:
            if key not in keys: break
        else: bases.append(PortTypeMap.context(PortTypeMap.ScheduledResourceTermination))

    #if base_notification is not None: 
    #    p = [\
    #        (base_notification,'Topic'), 
    #        (base_notification,'FixedTopicSet'), 
    #        (base_notification,'TopicExpressionDialects'),
    #    ]
    #    for key in p:
    #        if key not in keys: break
    #    else: bases.append('NotificationProducer')

    return bases


def GetPropertiesBasesFromRP(ged):
    '''
    NotificationProducerRP
    ScheduledResourceTerminationRP
    SubscriptionManagerRP
    '''
    assert ged.isElement() is True and ged.isLocal() is False,\
        'Unexpected parameter ged.'

    bases = []
    type_def = ged.getTypeDefinition()
    model_group = type_def.content
    extra_els = map(lambda er: er.getElementDeclaration(), model_group.content)

    namespaces = map(lambda e: e.getTargetNamespace(), extra_els)
    lifetime = None
    for nsuri in filter(lambda nsuri: nsuri in WSRF_V1_2.LIFETIME.XSD_LIST, namespaces):
        if lifetime is not None and lifetime != nsuri:
            warning.warn('Properties with different WS-ResourceLifetime namespaces, skipping properties generation.')
            return
        lifetime = nsuri

    base_notification = None
    for nsuri in filter(lambda nsuri: nsuri in WSRF_V1_2.BASENOTIFICATION.XSD_LIST, namespaces):
        if base_notification is not None and base_notification != nsuri:
            warning.warn('Properties with different WS-BaseNotification namespaces, skipping properties generation.')
            return
        base_notification = nsuri

    keys = map(lambda e: (e.getTargetNamespace(),e.getAttribute('name')), extra_els)
    if lifetime is not None and base_notification is not None: 
        p = [\
            (lifetime,'CurrentTime'),
            (lifetime,'TerminationTime'),
            (base_notification,'ConsumerReference'),
            (base_notification,'TopicExpression'),
            (base_notification,'UseNotify'),
            (base_notification,'Precondition'),
            (base_notification,'Selector'),
            (base_notification,'SubscriptionPolicy'),
            (base_notification,'CreationTime'),
        ]
        for key in p:
            if key not in keys: break
        else: bases.append('SubscriptionManagerRP')

    if lifetime is not None and 'SubscriptionManagerRP' not in bases:
        p = [\
            (lifetime,'CurrentTime'),
            (lifetime,'TerminationTime'),
        ]
        for key in p:
            if key not in keys: break
        else: bases.append('ScheduledResourceTerminationRP')

    if base_notification is not None: 
        p = [\
            (base_notification,'Topic'), 
            (base_notification,'FixedTopicSet'), 
            (base_notification,'TopicExpressionDialects'),
        ]
        for key in p:
            if key not in keys: break
        else: bases.append('NotificationProducerRP')

    return bases


class PortTypeMap:
    SubscriptionManager = 'SubscriptionManager'
    ScheduledResourceTermination = 'ScheduledResourceTermination'

    CONTEXT = {\
        SubscriptionManager:'SubscriptionManagerContext',
        ScheduledResourceTermination:'ScheduledResourceTerminationContext',
    }
    PROPERTIES = {\
        SubscriptionManager:'SubscriptionManagerRP',
        ScheduledResourceTermination:'ScheduledResourceTerminationRP',
    }
    __init__ = None

    def context(cls, key):
        return cls.CONTEXT.get(key)
    context = classmethod(context)

    def property(cls, key):
        return cls.PROPERTIES.get(key)
    property = classmethod(property)


class NameResolver:
    packages = []
    d = {\
        CLIENT:'pyGridWare.generated.stubs',
        SERVICE:'pyGridWare.generated.services',
        SERVICE_STUBS:'pyGridWare.generated.services.stubs',
        TYPES:'pyGridWare.generated.types',
        PROPERTIES:'pyGridWare.generated.properties',
        RESOURCE:'pyGridWare.generated.resource',
    }
    __init__ = None

    def Reset(cls, packages=None):
        cls.packages = packages or []
    Reset = classmethod(Reset)

    def NCName_to_Module(ncname):
        return TextProtect(ncname)
    NCName_to_Module = staticmethod(NCName_to_Module)

    def NCName_to_File(ncname):
        return TextProtect(ncname)
    NCName_to_File = staticmethod(NCName_to_File)

    def Module(cls, key, *args):
        s = cls.d[key]
        for a in map(cls.NCName_to_Module, args): s += '.%s' %a
        return s
    Module = classmethod(Module)

    def File(cls, key, *args):
        return '%s.py' %cls.Dir(key, *args)
    File = classmethod(File)

    def Dir(cls, key, *args):
        l = len(args)
        s = cls.d[key].replace('.', '/')
        for dir in map(cls.NCName_to_File, args): s += '/%s' %dir
        return s
    Dir = classmethod(Dir)

    # module names from section
    GetFileNameFromSection = staticmethod(lambda sec: '%s' %TextProtect(sec))
    GetContextModuleName = GetFileNameFromSection
    GetPropertiesModuleName = GetFileNameFromSection
    GetTypesModuleName = GetFileNameFromSection
    GetClientModuleName = GetFileNameFromSection
    GetServiceStubModuleName = GetFileNameFromSection
    GetServiceModuleName = GetFileNameFromSection

    GetModFromFile = staticmethod(lambda dir: dir.strip('/').replace('/','.').split('.py')[0])
    GetFileName = staticmethod(lambda dir,name: '%s/%s.py' %(dir.rstrip('/'), TextProtect(name)))

    GetPropertiesModuleFileName =\
        lambda cls, section, port:\
            cls.GetFileName(
                cls.MakeDirectories(cls.packages, cls.Dir(PROPERTIES), section, port.getService()),
                cls.GetPropertiesModuleName(port.name)
            )
    GetPropertiesModuleFileName = classmethod(GetPropertiesModuleFileName)

    GetContextModuleFileName =\
        lambda cls, section, service:\
            cls.GetFileName(
                cls.MakeDirectories(cls.packages, cls.Dir(RESOURCE), section),
                cls.GetContextModuleName(service.name)
            )
    GetContextModuleFileName = classmethod(GetContextModuleFileName)

    GetServiceModuleFileName =\
        lambda cls, section, service:\
            cls.GetFileName(
                cls.MakeDirectories(cls.packages, cls.Dir(SERVICE), section),
                cls.GetServiceModuleName(service.name)
            )
    GetServiceModuleFileName = classmethod(GetServiceModuleFileName)

    GetContextClassNameFromPort =\
        staticmethod(lambda port: '%sContext' %TextProtect(port.getPortType().name))

    GetPropertiesClassNameFromPort =\
        staticmethod(lambda port: '%s' %TextProtect(port.getPortType().name))


    def GetDirectoryPath(basedir, section, service=None):
        '''returns directory path.
        basedir  -- base directory, must exist.
        section  -- section
        service  -- WSDLTools service instance.
        '''
        BaseJoin = lambda basedir,dir: '%s/%s' %(basedir.rstrip('/'), TextProtect(dir))
        dir = BaseJoin(basedir, section)
        if service is not None:
            dir = BaseJoin(dir, service.name)
        return dir
    GetDirectoryPath = staticmethod(GetDirectoryPath)

    def MakeDirectories(cls, packages, basedir, section, service=None):
        '''uses path form GetDirectoryPath, creates any subdirs
        of basedir, adds subdirs to packages, and installs 
        an __init__ file in subdirs.
        packages -- list of packages to be added to distutils
        basedir  -- base directory, must exist.
        section  -- section
        service  -- WSDLTools service instance.
        '''
        assert type(packages) is list,\
            'Expecting list for parameter packages '
        assert os.path.isdir(basedir) is True,\
            'Expecting base directory for basedir not "%s"' %basedir
        assert service is None or isinstance(service, WSDLTools.Service) is True,\
            'Expecting service is WSDLTools.Service instance or None'

        AddPackage = lambda l,p: p in l or l.append(str(p))
        BaseJoin = lambda basedir,dir: '%s/%s' %(basedir.rstrip('/'), TextProtect(dir))
        #dirs = []
        #dir = secdir = BaseJoin(basedir, section)

        #AddPackage(packages, cls.GetModFromFile(secdir))
        #if service is not None:
        #    dir = BaseJoin(secdir, service.name)
        #    dirs.append(dir)
        #    AddPackage(packages, cls.GetModFromFile(dir))
        path = cls.GetDirectoryPath(basedir, section, service)
        assert path.startswith(basedir),\
            'directory path(%s) returned by GetDirectoryPath must start with basedir(%s)'\
            %(path, basedir)

        dir = basedir
        for d in path.split(basedir, 1)[-1].split('/'):
            if not d: continue
            dir = BaseJoin(dir, d)
            AddPackage(packages, dir)
            if os.path.isdir(dir) is False:
                assert os.path.exists(dir) is False, 'Expecting "%s" to be a directory' %dir
                os.makedirs(dir)

            fd = open(cls.GetFileName(dir,'__init__'), 'w+')
            fd.write(HEADER + '__all__=[]')
            fd.close()

        #return dir or secdir
        return dir

    MakeDirectories = classmethod(MakeDirectories)

