# System imports
import os, warnings, socket
from OpenSSL import SSL
from twisted.internet import tcp, main, interfaces, ssl
from twisted.internet import interfaces
from twisted.application.internet import _AbstractServer

# System imports
import socket

# Twisted imports
from twisted.python import log
from twisted.internet import base, address


def client_verify_cb(conn, cert, errnum, depth, ok):
    log.debug('Client Got certificate: %s' %cert.get_subject())
    for i in ('conn', 'cert', 'errnum', 'depth', 'ok'):
       log.debug("%s -- %s" %(i, eval(i)))
    return ok


def server_verify_cb(conn, cert, errnum, depth, ok):
    log.debug('Server Got certificate: %s' % cert.get_subject())
    for i in ('conn', 'cert', 'errnum', 'depth', 'ok'):
       log.debug("%s -- %s" %(i, eval(i)))
    return ok

def host_authz_cb(conn, cert, errnum, depth, ok):
    '''no good way to grab the CN=host/fqdn
    '''
    log.debug('Host Authz Got certificate: %s' %cert.get_subject())
    x509name = cert.get_subject()
    subject = str(x509name).split('\'')[1]
    if depth == 0:
        x509name = cert.get_subject()
        subject = str(x509name).split('\'')[1]
        peername = conn.getpeername()
        hostsuffix = '/CN=host/%s' %socket.getfqdn(peername[0])
        if subject.endswith(hostsuffix) is False: 
            warnings.warn(\
                'Host authorization failed: expected "%s", peer returned "%s"' \
                %(hostsuffix,subject)
            )
            return 0
    return ok


class ClientContextFactory(ssl.DefaultOpenSSLContextFactory):
    '''
    '''
    def __init__(self, privateKeyFileName, certificateFileName,
                 sslmethod=SSL.SSLv23_METHOD, calist=[], 
                 verify=client_verify_cb):
        self._calist = calist
        self._verify_cb = verify
        ssl.DefaultOpenSSLContextFactory.__init__(\
            self, privateKeyFileName, certificateFileName, sslmethod=sslmethod
        )

    def cacheContext(self):
        ssl.DefaultOpenSSLContextFactory.cacheContext(self)
        ctx = self.getContext()
        ctx.set_verify(SSL.VERIFY_PEER, self._verify_cb)
        for cert in self._calist:
            ctx.load_verify_locations(cert)


class ServerContextFactory(ssl.DefaultOpenSSLContextFactory):
    '''
    '''
    increment = 0

    def __init__(self, privateKeyFileName, certificateFileName,
                 sslmethod=SSL.SSLv23_METHOD, calist=[]):
        self._calist = calist
        ssl.DefaultOpenSSLContextFactory.__init__(\
            self, privateKeyFileName, certificateFileName, sslmethod=sslmethod
        )

    def cacheContext(self):
        ctx = SSL.Context(self.sslmethod)
        ctx.set_verify(SSL.VERIFY_PEER|SSL.VERIFY_FAIL_IF_NO_PEER_CERT, server_verify_cb)
        ctx.use_certificate_file(self.certificateFileName)
        ctx.use_privatekey_file(self.privateKeyFileName)
        for cert in self._calist:
            ctx.load_verify_locations(cert)

        #ctx.set_session_id("takes a string no longer than 32")
        ServerContextFactory.increment += 1
        ctx.set_session_id('%d' %ServerContextFactory.increment)
        self._context = ctx
   


class ACLServerContextFactory(ssl.DefaultOpenSSLContextFactory):
    '''
    '''
    increment = 0

    def __init__(self, privateKeyFileName, certificateFileName,
                 sslmethod=SSL.SSLv23_METHOD, calist=[], aclfile=None):
        '''
        file -- ACL file path
        '''
        self._aclfile = aclfile
        self._calist = calist
        ssl.DefaultOpenSSLContextFactory.__init__(\
            self, privateKeyFileName, certificateFileName, sslmethod=sslmethod
        )

    def getVerifyCB(self):
        '''
        '''
        if self._aclfile is None or os.path.isfile(self._aclfile) is False:
            raise RuntimeError, 'ACL File must be a file'

        subjects = map(lambda s: s.strip(), open(self._aclfile, 'r'))

        def verify_cb(conn, cert, errnum, depth, ok):
            # JRB no good way to extract a string, 
            # duplicate attributes not available via getattr
            x509name = cert.get_subject()
            subject = str(x509name).split('\'')[1]
            log.debug('ACL Authorization Server Got certificate: %s' % subject)
            for i in ('conn', 'cert', 'errnum', 'depth', 'ok'):
                log.debug("%s -- %s" %(i, eval(i)))
            if depth == 0 and subject not in subjects:
                warnings.warn("subject(%s) failed ACL authorization" %subject)
                return 0
            return ok

        return verify_cb

    def cacheContext(self):
        ctx = SSL.Context(self.sslmethod)
        ctx.set_verify(SSL.VERIFY_PEER|SSL.VERIFY_FAIL_IF_NO_PEER_CERT, self.getVerifyCB())
        ctx.use_certificate_file(self.certificateFileName)
        ctx.use_privatekey_file(self.privateKeyFileName)
        for cert in self._calist:
            ctx.load_verify_locations(cert)

        #ctx.set_session_id("takes a string no longer than 32")
        ServerContextFactory.increment += 1
        ctx.set_session_id('%d' %ServerContextFactory.increment)
        self._context = ctx
   


class GridMapServerContextFactory(ACLServerContextFactory):
    def getVerifyCB(self):
        '''
        '''
        if self._aclfile is None or os.path.isfile(self._aclfile) is False:
            raise RuntimeError, 'GridMap File must be a file'

        gridmap = dict(map(lambda s: s.strip('"').split('" ')[:2], open(self._aclfile, 'r')))
        subjects = gridmap.keys()

        def verify_cb(conn, cert, errnum, depth, ok):
            # JRB no good way to extract a string, 
            # duplicate attributes not available via getattr
            x509name = cert.get_subject()
            subject = str(x509name).split('\'')[1]
            log.debug('GridMap Authorization Server Got certificate: %s' % subject)
            for i in ('conn', 'cert', 'errnum', 'depth', 'ok'):
                log.debug("%s -- %s" %(i, eval(i)))
            if depth == 0 and subject not in subjects:
                warnings.warn("subject(%s) failed GridMap authorization (%s)" %(subject, subjects))
                return 0
            return ok

        return verify_cb

