##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, sys, warnings
from distutils import log
from distutils.core import setup
from distutils.dist import Distribution
from distutils.command.clean import clean as _clean
from distutils.command.build import build as _build
from distutils.dir_util import remove_tree
from os.path import isfile
from ConfigParser import ConfigParser
from ZSI.wstools.WSDLTools import WSDLReader

sys.path.append('%s/pyGridWare/utility/generate' %os.getcwd())
from Modules import WriteServiceModules, WriteClientStubModule,\
    WriteTypesModule

CONFIG_FILE = 'config.txt'
CACHE_FILE = '.pyGridWare.cache'


def RegenerateModules(items, services=True, stubs=True):
    '''Regenerate generated modules with new genrator
    '''
    if services is stubs is False:
        return

    add_packages = []
    for sec,url in items:
        log.info("Generate: %s, %s",sec,url)
        wsdl = None
        reader = WSDLReader()
        if isfile(url) is False:
            wsdl = reader.loadFromURL(url)
        else:
            wsdl = reader.loadFromFile(url)

        if services:
            WriteServiceModules(sec, wsdl, add_packages)
        if stubs:
            WriteClientStubModule(sec, wsdl)
        else:
            WriteTypesModule(sec, wsdl)

    return add_packages


class build(_build):
    '''Specialized build.
    '''
    user_options = _build.user_options[:]
    user_options.append(\
        ('generated', 'g',
         'create generated modules/directories and build cache file "%s"' %CACHE_FILE)
    )

    boolean_options = _build.boolean_options[:]
    boolean_options.append('generated')

    def initialize_options(self):
        _build.initialize_options(self)
        self.generated = None

    def run(self):
        '''if generate is specified generate modules.
        '''
        if not self.generated:
            return _build.run(self)

        cp = ConfigParser()
        cp.read(CONFIG_FILE)
        packages = RegenerateModules(cp.items('WSDL'), services=True, stubs=True)
        self.distribution.packages += packages
        ok = _build.run(self)
        fd = open(CACHE_FILE, 'w+')
        fd.write('[distribution]\npackages = %s' %packages)
        del cp
        return ok


class clean(_clean):
    '''Specialized clean.
    '''
    user_options = _clean.user_options[:]
    user_options.append(\
        ('generated', 'g',
         'remove all generated modules/directories')
    )

    boolean_options = _clean.boolean_options[:]
    boolean_options.append('generated')

    def initialize_options(self):
        _clean.initialize_options(self)
        self.generated = None

    def run(self):
        '''if generated is specified remove all such files.
        '''
        _clean.run(self)
        if not self.generated: return

        func1 = lambda f: f != '__init__.py' and f.endswith('.py')
        files = []
        for dir in ('pyGridWare/generated/stubs/','pyGridWare/generated/types/','pyGridWare/generated/services/stubs/'):
            for f in map(lambda f: os.path.join(dir,f), filter(func1, os.listdir(dir))):
                log.info("Remove: %s", f)
                os.remove(f)

        if os.path.isfile(CACHE_FILE):
            cp = ConfigParser()
            cp.read(CACHE_FILE)
            packages = eval(cp.get('distribution', 'packages'))
            for dir in packages:
                if dir.startswith('pyGridWare/generated') is False:
                    log.warn("'%s' is not in pyGridWare/generated -- can't remove it", dir)
                if os.path.exists(dir):
                    remove_tree(dir, dry_run=self.dry_run)
                else:
                    log.warn("'%s' does not exist -- can't remove it", dir)

            del cp
            os.remove(CACHE_FILE)


class PyGridWareDist(Distribution): 
    global_options = Distribution.global_options[:]
    global_options.append(('regenerate', None, "Deprecated, use build command with generated flag"))

    def parse_command_line(self):
        ok = Distribution.parse_command_line(self)
        if self.verbose == 1:
            from ZSI.wstools.logging import setBasicLoggerDEBUG
            setBasicLoggerDEBUG()

        if '--regenerate' in self.script_args:
            raise DeprecationWarning, \
                'regenerate global option no longer supported.  Use build command.'

        cp = ConfigParser()
        if os.path.isfile(CACHE_FILE) is True:
            cp = ConfigParser()
            cp.read(CACHE_FILE)
            self.packages += eval(cp.get('distribution', 'packages'))
            del cp
        else:
            warnings.warn(\
                '***No Modules have been generated ("%s")\n\tpython setup.py build -g' %CACHE_FILE,
            )
        return ok 


def main():
    classifiers = """\
        Development Status :: 5 - Production/Stable
        Intended Audience :: Developers
        Intended Audience :: Science/Research
        License :: OSI Approved :: BSD License
        Programming Language :: Python
        Natural Language :: English
        Operating System :: Microsoft :: Windows
        Operating System :: POSIX
        """

    longDesc = """
        Python implementation of the WS-Resource Framework (WSRF).  
        pyGridWare has facilities for secure messaging including security 
        context negotiation, delegation, and GSS XML Signatures.
        """

    setup(name="pyGridWare", 
          version="1.0.2beta", 
          packages=["pyGridWare", 
                    "pyGridWare.addressing", 
                    "pyGridWare.generated", 
                    "pyGridWare.generated.properties", 
                    "pyGridWare.generated.resource", 
                    "pyGridWare.generated.services", 
                    "pyGridWare.generated.services.stubs", 
                    "pyGridWare.generated.stubs", 
                    "pyGridWare.generated.types", 
                    "pyGridWare.lifetime", 
                    "pyGridWare.notification", 
                    "pyGridWare.properties",
                    "pyGridWare.faults",
                    "pyGridWare.resource", 
                    "pyGridWare.security", 
                    "pyGridWare.security.gss", 
                    "pyGridWare.security.base",
                    "pyGridWare.services", 
                    "pyGridWare.utility", 
                    "pyGridWare.utility.generate", 
                    "pyGridWare.utility.scripts",
                    "pyGridWare.wsrf", 
          ],
          scripts=["tools/WSDLBuilder"],
          description="Python WS-Resource Framework and GSS Security",
	  author="Joshua Boverhof", 
          author_email="JRBoverhof@lbl.gov",
          maintainer="Joshua Boverhof", 
          maintainer_email="JRBoverhof@lbl.gov",
	  url="http://www-itg.lbl.gov/gtg/pyGridWare/",
          classifiers = filter(None, classifiers.split("\n"))[:-1],
          long_description=longDesc,
          distclass=PyGridWareDist,
          cmdclass={'clean':clean, 'build':build},
    )


if __name__ == '__main__': 
    main()
